# -*- coding: utf-8 -*-
# Original code by (C) 2019 yorikvanhavre <yorik@uncreated.net>
# Copyright (C) 2023 Ilia Kurochkin <brothermechanic@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
'''
DESCRIPTION.
Main module:
- Reads a FreeCAD .FCStd file.
- Set tesselation parts to mesh.
- Inport meshes in Blender scene.
- Setup hierarchy.
- Setup materials.
- Setup LCS points.
- Apply FreeCAD to Bledner scene transforms.
'''
__version__ = '0.3'
import time
import FreeCAD
import Part, Mesh, MeshPart
import logging
import math
import xml
import sys
import xml.sax
import zipfile
import os
import random
import bpy
from import_fcstd.import_hierarchy import (hierarchy,
                                           placement)
from import_fcstd.import_materials import (assign_materials,
                                           assign_black)
from import_fcstd.is_object_solid import is_object_solid
from utils.object_transforms import apply_transforms
logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


# COLLECTIONS NAMIG CONVENTION
parts_col_name = 'Import Parts'
lcs_col_name = 'Import LCS'
hierarchy_col_name = 'Import Hierarchy'
lowpoly_col_name = 'Lowpoly Parts'
# LCS POINT'S SUFFIXES CONVENTION
inlet = '_in'
outlet = '_out'
root = '_root'
# CG ASSETS SUFFIXES CONVENTION
hightpoly = '_hp'
lowpoly = '_lp'
render = '_render'


def obj_importer(filename,
                 tesselation_method='Standard',
                 linear_deflection=0.1,
                 angular_deflection=30.0,
                 fem_size=5.0,
                 update=False,
                 scene_placement=True,
                 skiphidden=True,
                 scale=0.001,
                 select=True,
                 nonsolid_property='Robossembler_NonSolid'):

    ''' Reads a FreeCAD .FCStd file and creates Blender objects '''

    doc = FreeCAD.open(filename)
    docname = doc.Name

    if not update:
        parts_collection = bpy.data.collections.new(parts_col_name)
        bpy.context.scene.collection.children.link(parts_collection)

        lcs_collection = bpy.data.collections.new(lcs_col_name)
        bpy.context.scene.collection.children.link(lcs_collection)

        hierarchy_collection = bpy.data.collections.new(hierarchy_col_name)
        bpy.context.scene.collection.children.link(hierarchy_collection)

    # collect all materials
    fem_mats = []
    for fem_mat in doc.Objects:
        if fem_mat.isDerivedFrom('App::MaterialObjectPython'):
            fem_mats.append(fem_mat)
    bobjs = []
    bobjs_for_render = []
    for obj in doc.Objects:
        bobj = None
        if skiphidden:
            if not obj.Visibility:
                continue

        if obj.isDerivedFrom('PartDesign::CoordinateSystem'):
            if update:
                # locate existing object
                for o in bpy.data.objects:
                    if o.name == obj.Label:
                        bobj = o
                        logger.debug('Replacing existing %s', obj.Label)
            else:
                bobj = bpy.data.objects.new(obj.Label, None)
                bobj.empty_display_type = 'ARROWS'
                bobj.empty_display_size = round(random.uniform(0.05, 0.15), 3)
                bobj.show_in_front = True
                lcs_collection.objects.link(bobj)

        elif obj.isDerivedFrom('Part::Feature'):
            # filter for nonsolids
            if is_object_solid(obj) or hasattr(obj, nonsolid_property):
                verts = []
                edges = []
                faces = []
                # create mesh from shape
                shape = obj.Shape
                if scene_placement:
                    shape = obj.Shape.copy()
                    shape.Placement = obj.Placement.inverse().multiply(shape.Placement)
                meshfromshape = doc.addObject('Mesh::Feature','Mesh')
                if tesselation_method == 'Standard':
                    meshfromshape.Mesh = MeshPart.meshFromShape(
                        Shape=shape,
                        LinearDeflection=linear_deflection,
                        AngularDeflection=math.radians(angular_deflection),
                        Relative=False)
                elif tesselation_method == 'FEM':
                    meshfromshape.Mesh = MeshPart.meshFromShape(
                        Shape=shape,
                        MaxLength=fem_size)
                else:
                    raise TypeError('Wrong tesselation method! '
                                    'Standard and FEM methods are supported only!')
                    break
                t = meshfromshape.Mesh.Topology
                verts = [[v.x,v.y,v.z] for v in t[0]]
                faces = t[1]

                if verts and faces:
                    # create or update object with mesh and material data
                    bmesh = bpy.data.meshes.new(name=obj.Label)
                    bmesh.from_pydata(verts, edges, faces)
                    bmesh.update()
                    if update:
                        # locate existing object
                        for o in bpy.data.objects:
                            if o.name == obj.Label:
                                bobj = o
                                bobj.data = bmesh
                                logger.debug('Replacing existing %s', obj.Label)
                    else:
                        bobj = bpy.data.objects.new(obj.Label, bmesh)
                        parts_collection.objects.link(bobj)

        # skip for other object's types
        else:
            continue

        if bobj and scene_placement and not update:
            placement(bobj, obj, scale)
            if bobj.type == 'MESH':
                bobj.scale = (scale, scale, scale)
                apply_transforms(bobj, scale=True)

            # construct assembly hierarchy
            hierarchy_objs = hierarchy(bobj, obj, scale)
            for hierarchy_obj in hierarchy_objs:
                hierarchy_collection.objects.link(hierarchy_obj)

            # one material for the whole object
            if bobj.type == 'MESH':
                for fem_mat in fem_mats:
                    for ref in fem_mat.References:
                        if ref[0].Label == bobj.name:
                            assign_materials(bobj, fem_mat)
                            bobjs_for_render.append(bobj)
                            continue
                # looks like this is hidden internal object
                if not bobj.material_slots:
                    assign_black(bobj)

        # optional select object after importing
        if bobj and select:
            bpy.context.view_layer.objects.active = bobj
            bobj.select_set(True)

        bobjs.append(bobj)

    # losted root lcs inlet workaround
    lcs_objects = lcs_collection.objects
    if lcs_objects:
        root_lcs = [lcs for lcs in lcs_objects if lcs.name.endswith(root)][0]
        root_inlet_name = ('{}{}'.format(root_lcs.name.split(root)[0], inlet))
        if not bpy.data.objects.get(root_inlet_name):
            root_inlet = bpy.data.objects.new(root_inlet_name, None)
            root_inlet.empty_display_type = 'ARROWS'
            root_inlet.empty_display_size = 0.1
            root_inlet.show_in_front = True
            root_inlet.location = root_lcs.location
            root_inlet.rotation_euler = root_lcs.rotation_euler
            root_inlet.parent = root_lcs.parent
            lcs_collection.objects.link(root_inlet)

    FreeCAD.closeDocument(docname)

    # TODO
    # update do not dork
    logger.info('Imported %s objects without errors', len(bobjs))
    return bobjs_for_render
